/* -*- mode: c++; c-basic-offset: 3; indent-tabs-mode: nil; -*- */
// System Header Files   
#include <sys/stat.h>

#include <cstring>   
#include <stdexcept>   
#include <cstdio>   
#include <unistd.h>   
#include <cerrno>

#include <sstream>
   
// General Header Files   
#include "general/autoarray.hh"
#include "general/Thread.hh"
#include "general/util.hh"

#include "genericAPI/Logging.hh"
#include "genericAPI/Stat.hh"

// Local Header Files
#include "cachefile.hh"   
#include "DumpCacheDaemon.hh"
#include "mountpt.hh"   
#include "timehash.hh"   

using namespace diskCache;
// using namespace General;   
// using namespace std;

using General::ReadWriteLock;
using General::AutoArray;

using GenericAPI::LogEntryGroup_type;
using GenericAPI::queueLogEntry;

using std::exception;
using std::ios_base;
using std::list;
using std::string;
using std::streamsize;
using std::runtime_error;
using std::vector;

using General::ifstream;

/*! \file
 *
 *  \brief The definition of diskCache::CacheFile.
 */

// Local data initialization.   
namespace 
{
   const size_t stream_buffer_size( 1024 * 1024 );

   class cacheHeader
   {
   public:
      typedef INT_2U version_type;

      cacheHeader( version_type Version
                   = diskCache::CacheFile::CURRENT_VERSION );

      bool NeedToSwapData( ) const;
      void Read( std::istream& Stream );
      void Write( std::ostream& Stream ) const;
      version_type Version( ) const;

   private:
      static const INT_4U       SWAP_REFERENCE;
      static const CHAR         TAG[];
      static const CHAR         TAG_OLD[];

      version_type      m_version;
      INT_4U            m_swap_data;
   };

   inline cacheHeader::version_type cacheHeader::
   Version( ) const
   {
      return m_version;
   }

   // Default file name to use if Tcl doesn't provide one
   static const CHAR* const DefaultFileName( ".frame.cache2" );
   
   static const string TmpExtension( ".tmp" );
   static const string BackupExtension( ".bak" );   
   
   static const streamsize size_t_bytes( sizeof( size_t ) );
   static const streamsize char_bytes( sizeof( CHAR ) );
   static const streamsize int_4s_bytes( sizeof( INT_4S ) );
   static const streamsize int_4u_bytes( sizeof( INT_4U ) );   
   static const streamsize int_8u_bytes( sizeof( INT_8U ) );   
   
   
   const INT_4U cacheHeader::SWAP_REFERENCE = 0x12345678;
   const CHAR cacheHeader::TAG[] = "<DiSKCaCHeVerSIONed>";
   const CHAR cacheHeader::TAG_OLD[] = "<DiSKCaCHe>";

   cacheHeader::
   cacheHeader( version_type Version )
      : m_version( Version ),
        m_swap_data( SWAP_REFERENCE )
   {
   }

   inline bool cacheHeader::
   NeedToSwapData( ) const
   {
      return SWAP_REFERENCE != m_swap_data;
   }

   void cacheHeader::
   Read( std::istream& Stream )
   {
      static const int tag_size = sizeof( TAG ) - 1;
      static const int tag_old_size = sizeof( TAG_OLD ) - 1;

      CHAR      tag_buffer[ tag_size ];

      //-----------------------------------------------------------------
      // Check for version 0x0000 (pre-header days)
      //-----------------------------------------------------------------
      Stream.read( tag_buffer, tag_old_size );
      if ( ::strncmp( TAG_OLD, tag_buffer, tag_old_size ) == 0 )
      {
         m_version = 0x0000;
         return;
      }
      //-----------------------------------------------------------------
      // Check for versions which have defined headers
      //-----------------------------------------------------------------
      Stream.read( &tag_buffer[ tag_old_size ], tag_size - tag_old_size );
      if ( strncmp( TAG, tag_buffer, tag_size ) != 0 )
      {
         throw runtime_error( "Not a frame cache file" );
      }
      //-----------------------------------------------------------------
      // It is a frame cache file that has header information
      //-----------------------------------------------------------------
      // Version 0x0100
      Stream.read( static_cast< CHAR *>( (void*)( &m_swap_data ) ),
                   streamsize( sizeof( m_swap_data ) ) );
      Stream.read( static_cast< CHAR * >( (void*)( &m_version ) ),
                   streamsize( sizeof( m_version ) ) );
      if ( m_swap_data != SWAP_REFERENCE )
      {
         INT_4U data = m_swap_data;
         reverse< sizeof( data ) >( &data, 1 );
         if ( data != SWAP_REFERENCE )
         {
            throw runtime_error( "Unable to determine byte ordering" );
         }
      }
      if ( NeedToSwapData( ) )
      {
         reverse< sizeof( m_version ) >( &m_version, 1 );
      }
   }

   void cacheHeader::
   Write( std::ostream& Stream ) const
   {
      Stream.write( TAG, sizeof( TAG ) - 1 );
      // Added with version 0x0100
      Stream.write( static_cast< const CHAR* >( (const void*)( &m_swap_data ) ),
                    streamsize( sizeof( m_swap_data ) ) );
      Stream.write( static_cast< const CHAR* >( (const void*)( &m_version ) ),
                    streamsize( sizeof( m_version ) ) );
   }
}
   

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
const INT_2U diskCache::CacheFile::CURRENT_VERSION = 0x0100;

void diskCache::CacheFile::
ValidateVersion( version_type Version )
{
   switch( Version )
   {
   case 0x0000: // Origional version
   case 0x0100: // Added byte swapping support
   case 0x0101: // Added support for extentions other than .gwf
      break;
   default:
      {
         std::ostringstream     msg;

         msg << "Unsupported version: " << Version
            ;
         throw std::range_error( msg.str( ) );
      }
      break;
   }
}
//-----------------------------------------------------------------------
/// This method extracts next available input character, but does not
/// actually remove the character from the input stream.
///
/// \return std::ios_base::int_type Next available character, otherwise,
///     it returns traits_type::eof.  
//-----------------------------------------------------------------------
std::ios::int_type diskCache::CacheFile::
peek()
{
   return mStream.peek();
}
   
   
//-----------------------------------------------------------------------
/// This method writes global frame hash to the binary file in the current 
/// directory. Tcl layer can specify a different file name for each "write"
/// operation. If "tcl_file_name" already exists it will be saved as a
/// "tcl_file_name.bak" file.
/// Method will use default file name in case Tcl doesn't provide one.
///   
/// \param[in] tcl_file_name A file name to write frame hash to.
///
/// \return Nothing.
///
/// \exception runtime_error Could not open file "name" for writing.
/// \exception runtime_error Could not remove old backup copy of cache file.
/// \exception runtime_error Could not create backup copy of already
///     existing cache file.
/// \exception runtime_error Could not move temporary copy of cache file
///     to destination.
//-----------------------------------------------------------------------
void diskCache::CacheFile::
writeToFile( const CHAR* const tcl_file_name,
             mode_type Mode,
             version_type Version )
{
   using General::Thread;

   // Thread::Self::CancellationType( Thread::CANCEL_EXCEPTION );

   static const string method_name( "CacheFile::writeToFile(): " );

   
   string file_name( DefaultFileName );
   
   // Tcl provided file name
   if( tcl_file_name && *tcl_file_name != '\0' )
   {
      file_name = tcl_file_name;
   }
   
   
   // Make sure stream is closed.
   if( mStream.is_open() )
   {
      mStream.close();
   }

   
   // At first, write to the temporary file in current directory,
   // then move it to the destination location: "file_name" in the current directory.
   string tmp_file_name( file_name );
   tmp_file_name += TmpExtension;
   
   if ( m_stream_buffer.get( ) == NULL )
   {
      m_stream_buffer.reset( new char[ stream_buffer_size ] );
   }
   if ( m_stream_buffer.get( ) )
   {
      mStream.rdbuf( )->pubsetbuf( m_stream_buffer.get( ), stream_buffer_size );
   }
   mStream.clear();   
   mStream.open( tmp_file_name.c_str(), ios_base::out | ios_base::binary );
   
   if( !mStream )
   {
      string msg( method_name );
      msg += "Could not open file \'";
      msg += tmp_file_name;
      msg += "\' for writing.";
   
      throw runtime_error( msg.c_str() );
   }

   
   try
   {
      // Cause stream failures to throw an exception
      mStream.clear();   
      mStream.exceptions( ios_base::badbit | ios_base::failbit );

   
      // Write file header
      cacheHeader       header( Version );

      m_version = header.Version( );
      header.Write( mStream );
   
      // Write global frame hash   
      MountPointHash::write( *this );
   
   
      // Reset exception mask: don't throw exceptions any more
      // Clear stream state
      mStream.clear();
      mStream.exceptions( ios_base::goodbit );
   }
   catch( const exception& exc )
   {
      // Clear stream state
      mStream.clear();
      // Reset exception mask: don't throw exceptions any more
      mStream.exceptions( ios_base::goodbit );

   
      // Close the file stream
      mStream.close();
   
      // Remove file
      unlink( tmp_file_name.c_str() );
    
   
      string msg( method_name );
      msg += exc.what();
   
      throw runtime_error( msg.c_str() );
   }
   
   
   mStream.close();
   mStream.clear();   
   
   
   // Create backup copy of existing frame cache file
   ifstream cache_file( file_name.c_str() );
   
   if( cache_file )
   {
      // There is existing cache file ---> need to save it
      cache_file.close();

   
      // Format name for backup file
      string backup_name( file_name );
      backup_name += BackupExtension;

   
      // Check if there is a backup file from previous write:
      ifstream backup_file( backup_name.c_str() );
   
      if( backup_file )
      {
         backup_file.close();
   
         if( unlink( backup_name.c_str() ) )
         {
            // Could not remove old backup file
            const string errno_msg( strerror( errno ) );
      
            string msg( method_name );
            msg += "Could not remove old backup copy of cache file \'";
            msg += backup_name;
            msg += "\' (errno=";
            msg += errno_msg;
            msg += "). New cache file is stored as temporary \'";
            msg += tmp_file_name;
            msg += '\'';
      
            throw runtime_error( msg.c_str() );   
         }
      }
   
   
      if( rename( file_name.c_str(), backup_name.c_str() ) )
      {
         // Could not rename file
         const string errno_msg( strerror( errno ) );
   
         // Creating backup copy failed:
         string msg( method_name );
         msg += "Could not create backup copy of already existing cache file \'";
         msg += file_name;
         msg += "\' (errno=";
         msg += errno_msg;
         msg += "). New cache file is stored as temporary \'";
         msg += tmp_file_name;
         msg += '\'';
   
         throw runtime_error( msg.c_str() );
      }
   }

   
   // Move file to the requested destination
   if( rename( tmp_file_name.c_str(), file_name.c_str() ) )
   {
      // Could not rename file
      const string errno_msg( strerror( errno ) );
   
      // Creating backup copy failed:
      string msg( method_name );
      msg += "Could not move temporary copy of cache file \'";
      msg += tmp_file_name;
      msg += "\' (errno=";
      msg += errno_msg;
      msg += ") to the destination \'";
      msg += file_name;
      msg += '\'';
   
      throw runtime_error( msg.c_str() );   
   }

   
   return;
}
   
   
//-----------------------------------------------------------------------
/// This method reads frame hash from the binary file.
/// Tcl layer can specify a different file name for each "read" operation.
///   
/// Method will use default file name in case Tcl doesn't provide one.   
///   
/// \param[in] tcl_file_name A file name to read frame hash from.
///
/// \return Nothing.
///
/// \exception runtime_error Could not open file "name" for reading.
//-----------------------------------------------------------------------
void diskCache::CacheFile::
readFromFile( const CHAR* const tcl_file_name )
{
   static const char method_name[] = "diskCache::CacheFile::readFromFile(): ";

   try
   {


      string file_name( DefaultFileName );
   
      // Tcl provided file name
      if( tcl_file_name && *tcl_file_name != '\0' )
      {
         file_name = tcl_file_name;
      }
      
      //-------------------------------------------------------------------------
      // Determine if a temporary file should be read instead.
      //-------------------------------------------------------------------------
      try {
         struct stat reg;
         struct stat tmp;
         std::string    tmp_name( file_name );
         tmp_name += DumpCacheDaemon::RESYNC_EXTENSION;

         if ( GenericAPI::Stat( tmp_name, tmp ) == 0 )
         {
            //-------------------------------------------------------------------
            // The temporary files exists.
            // Next:
            //  1) Check for the existance of the regular file.
            //  2) If the regular file does exist, then see which file is more
            //     recent.
            //-------------------------------------------------------------------
            if ( ( ( GenericAPI::Stat( file_name, reg ) != 0 )
                   || ( reg.st_mtime < tmp.st_mtime ) ) )
            {
               //----------------------------------------------------------------
               // The resync filename is more recent which means that its data
               // more acurately reflects the last request made to the system
               // before shutdown.
               // Use it!
               //----------------------------------------------------------------
               file_name = tmp_name;
            }
         }
      }
      catch( ... )
      {
         //----------------------------------------------------------------------
         // Something went wrong, just use the filename supplied when called.
         //----------------------------------------------------------------------
      }
      
      if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
      {
         std::ostringstream        msg;

         msg << "Attempting to read cache file: " << file_name
            ;
         queueLogEntry( msg.str( ),
                        LogEntryGroup_type::MT_DEBUG,
                        30,
                        method_name,
                        "CXX_INTERFACE" );
      }

      // Make sure stream is closed.
      if( mStream.is_open() )
      {
         mStream.close();
      }

      if ( m_stream_buffer.get( ) == NULL )
      {
         m_stream_buffer.reset( new char[ stream_buffer_size ] );
         mStream.rdbuf( )->pubsetbuf( m_stream_buffer.get( ), stream_buffer_size );
      }
      mStream.clear();   
      mStream.open( file_name.c_str(), ios_base::in | ios_base::binary );
   
      if( !mStream )
      {
         string msg( method_name );
         msg += "Could not open file \'";
         msg += file_name;
         msg += "\' for reading.";
   
         throw runtime_error( msg.c_str() );
      }
   

      // Read frame hash
      try
      {   
         // Make sure stream state is clear
         mStream.clear();   
   
         // Cause stream failures to throw an exception
         mStream.exceptions( ios_base::badbit | ios_base::failbit );
   
   
         // Read file header
         {
            cacheHeader    header;
            header.Read( mStream );
            m_need_to_swap_data = header.NeedToSwapData( );
            m_version = header.Version( );
         }

   
         MountPointHash::read( *this );
   
   
         // Reset exception mask: don't throw exceptions any more
         mStream.clear();   
         mStream.exceptions( ios_base::goodbit );
      }
      catch( const exception& exc )
      {
         // Clear stream state
         mStream.clear();
         // Reset exception mask: don't throw exceptions any more
         mStream.exceptions( ios_base::goodbit );

   
         // Close the file stream
         mStream.close();

   
         string msg( method_name );
         msg += exc.what();
   
         throw runtime_error( msg.c_str() );
      }

   
      mStream.close();
      mStream.clear();   
   
   
      return;
   }
   catch( const std::exception& Exception )
   {
      std::ostringstream        msg;

      msg << "Failed to read file: "
          << " Exception: " << Exception.what( )
         ;
      queueLogEntry( msg.str( ),
                     LogEntryGroup_type::MT_WARN,
                     0,
                     method_name,
                     "CXX_INTERFACE" );
      throw;       // Rethrow the exception
   }
}   


/**
   \param[in] ptr A pointer to the buffer to write to the file.
   \param[in] size Number of bytes to write.   
*/
void diskCache::CacheFile::
write( const void* ptr, const std::streamsize& size )
{
   mStream.write( static_cast< const CHAR* >( ptr ), size );
   
   return;
}
   

/**
    \param[out] ptr A pointer to the buffer to read from the file to.
    \param[in] size Number of bytes to read.   
*/
void diskCache::CacheFile::
read( void* ptr, const std::streamsize& size )
{
   mStream.read( static_cast< CHAR* >( ptr ), size );
   
   return;
}

/**
   \page CacheFile

   \section OverviewCacheFormat Overview of the Disk Cache Formats

   The diskcacheAPI has several ways of presenting the in memory cache
   of frame files.
   The most easily accessable version of this information is the ascii
   representation available through the LDAS user command \e getFrameCache .
   This form is described in the section titled
   \ref CacheAsciiFormat "Ascii Cache File Format".
   The binary format is used by the disckcacheAPI at startup to seed the
   cache with previously discoverd information.
   This form is described in the section titled
   \ref CacheBinaryFormat "Binary Cache File Format".

   \anchor CacheAsciiFormat
   \section AsciiCacheFormat Ascii Cache File Format

   The ascii representation of the state of the in memory cache of files
   is a Tcl list. Each of the list items is composed of 4 elements.
   The first element is a composit of five values;
   the full path to the directory, the list of IFOs,
   the frame type,
   the number of frames per file,
   and the duration of the frames.
   The second element is an integer representing the time the directory was
   last modified.
   The third element specifies the number of frames in the directory.
   The fourth element is a list where the
   first element is the GPS start time of the first frame in the set
   and the second element is the GPS end time.

   \anchor CacheBinaryFormat
   \section BinaryCacheFormat Binary Cache File Format

   The graph below symbolizes the format of the memory cache image
   as written to media in binary form.
   For efficiency purposes, all numeric values are stored on media in
   the native form of the writing application.

   \dot
   digraph MediaImage {
     node [shape=record, fontname=Helvetica, fontsize=10 ];
     start [ label="start", shape="circle" ];
     stop [ label="End of File", shape="circle" ];
     tag_nver [ label="\<DiSKCaCHe\>" ];
     tag_ver [ label="\<DiSKCaCHeVerSIONed\>" ];
     hdr_0x0100 [ label="Version 0x0100 Header Block",
                  URL="\ref r_hdr_0x0100"
                ];
     mountPointBlock [ label="mountPointBlock"
                  URL="\ref r_mountPointBlock"
                ];
     excludeDirPatternBlock [ label="excludeDirPatternBlock"
                  URL="\ref r_excludeDirPatternBlock"
                ];
     cacheDirEntry [ label="cacheDirEntry"
                  URL="\ref r_cacheDirEntry"
                ];
     frameIFOAndType [ label="frameIFOAndType"
                  URL="\ref r_frameIFOAndType"
                ];
     Interval [ label="Interval"
                  URL="\ref r_Interval"
                ];

     start -> tag_nver [ arrowhead="open", style="dashed" ];
     start -> tag_ver [ arrowhead="open", style="dashed" ];
     tag_ver -> hdr_0x0100 [ arrowhead="open", style="dashed" ];
     tag_nver -> mountPointBlock [ arrowhead="open", style="dashed" ];
     hdr_0x0100 -> mountPointBlock [ arrowhead="open", style="dashed" ];
     mountPointBlock -> excludeDirPatternBlock [ arrowhead="open", style="dashed" ];
     excludeDirPatternBlock -> cacheDirEntry [ arrowhead="open", style="dashed" ];
     excludeDirPatternBlock -> stop [ arrowhead="open", style="dashed" ];
     cacheDirEntry -> cacheDirEntry [ arrowhead="open", style="dashed" ];
     cacheDirEntry -> frameIFOAndType [ arrowhead="open", style="dashed" ];
     frameIFOAndType -> Interval [ arrowhead="open", style="dashed" ];
     cacheDirEntry -> stop [ arrowhead="open", style="dashed" ];
   }
   \enddot

   Below is a table describing the basic data types that are written
   to media.

   <table>
   <caption>Basic data types as written to media</caption>
   <tr><td><b>Type</b></td><td><b>Description</b></td></tr>
   <tr><td>INT_2U</td><td>A 2 byte unsigned integer</td></tr>
   <tr><td>INT_4U</td><td>A 4 byte unsigned integer</td></tr>
   <tr><td>INT_4S</td><td>A 4 byte signed integer</td></tr>
   <tr><td>INT_8U</td><td>An 8 byte unsigned integer</td></tr>
   <tr><td>string</td><td>An INT_8U followed by a character array of the size of the INT_8U value</td></tr>
   </table>

   The binary cache file starts either with the ascii tag <b>\<DiSKCaCHe\></b>
   or the ascii tag <b>\<DiSKCaCHeVerSIONed\></b>.
   If the file starts with the ascii tag <b>\<DiSKCaCHeVerSIONed\></b>,
   then a version specific header block follows.

   \anchor r_hdr_0x0100
   <table>
   <caption>Version 0x0100 Header Block</caption>
   <tr><td><b>Type</b></td><td><b>Name</b></td><td><b>Description</b></td></tr>
   <tr><td>INT_4U</td><td>byteRef</td><td>Contains the hex value 0x12345678 as written by an <b>INT_4U</b></td></tr>
   <tr><td>INT_2U</td><td>version</td><td>Contains the version number of the cache file</td></tr>
   </table>

   The value of \e byteRef should be used to determine if any byte reordering
   needs to performed by the software reading the cache file. This information
   cannot be determined for files that start with the tag <b>\<DiSKCaCHe\></b>
   as they contain no header block.

   \anchor r_mountPointBlock
   \subsection mountpointlist Mount Point List

   <table>
   <caption>Definition of mountPointBlock</caption>
   <tr><td><b>Type</b></td><td><b>Name</b></td><td><b>Description</b></td></tr>
   <tr>
   <td>INT_8U</td>
   <td>nMountPoints</td>
   <td>Number of mount points</td>
   </tr>
   <tr>
   <td>string[nMountPoints]</td>
   <td>mountPoints</td>
   <td>An array of size \e nMountPoints of mount points</td>
   </tr>
   </table>
   

   \anchor r_excludeDirPatternBlock
   \subsection excludedirlist List of directory patterns to exclude

   <table>
   <caption>Definition of excludeDirPatternBlock</caption>
   <tr><td><b>Type</b></td><td><b>Name</b></td><td><b>Description</b></td></tr>
   <tr>
   <td>INT_8U</td>
   <td>nExclude</td>
   <td>Number of directory exclude patterns</td>
   </tr>
   <tr>
   <td>string[nExclude]</td>
   <td>patterns</td>
   <td>An array of size \e nExclude of patterns of directories to exclude</td>
   </tr>
   </table>
   
   \anchor r_cacheDirEntry
   \subsection directories Cached Directory Information

   <table>
   <caption>Definition of cacheDirEntry</caption>
   <tr><td><b>Type</b></td><td><b>Name</b></td><td><b>Description</b></td></tr>
   <tr>
   <td>string</td>
   <td>dir</td>
   <td>Directory path relative to \e mountpt</td>
   </tr>
   <tr>
   <td>string</td>
   <td>mountpt</td>
   <td>Fully qualified directory path to the mount point</td>
   </tr>
   <tr>
   <td>INT_8U</td>
   <td>nSubDirs</td>
   <td>Number of subdirectories in the current directory</td>
   </tr>
   <tr>
   <td>string[nSubDirs]</td>
   <td>subDirs</td>
   <td>Array of subdirectory names</td>
   </tr>
   <tr>
   <td>INT_8U</td>
   <td>nExcludedSubDirs</td>
   <td>Number of excluded subdirectories in the current directory</td>
   </tr>
   <tr>
   <td>string[nExcludedSubDirs]</td>
   <td>excludedSubDirs</td>
   <td>Array of excluded subdirectory names</td>
   </tr>
   <tr>
   <td>INT_4S</td>
   <td>mTime</td>
   <td>The last time \e dir was modified</td>
   </tr>
   <tr>
   <td>INT_8U</td>
   <td>nElements</td>
   <td>Number of cached directory elements</td>
   </tr>
   <tr>
   <td>frameIFOAndType[nElements]</td>
   <td>frameEntries</td>
   <td>Array containing information about the frame files within the \e dir</td>
   </tr>
   </table>

   \anchor r_frameIFOAndType
   <table>
   <caption>Definition of frameIFOAndType</caption>
   <tr><td><b>Type</b></td><td><b>Name</b></td><td><b>Description</b></td></tr>
   <tr>
   <td>string</td>
   <td>ifoAndType</td>
   <td>IFO-Type- formatted string</td>
   </tr>
   <tr>
   <td>INT_4U</td>
   <td>nElements</td>
   <td>Number of elements</td>
   </tr>
   <tr>
   <td>Interval[nElements]</td>
   <td>intervals</td>
   <td>Array of intervals</td>
   </tr>
   </table>

   \anchor r_Interval
   <table>
   <caption>Definition of Interval</caption>
   <tr><td><b>Type</b></td><td><b>Name</b></td><td><b>Description</b></td></tr>
   <tr><td>INT_4U</td><td>start</td><td>Start time of the interval in seconds</td>
   <tr><td>INT_4U</td><td>stop</td><td>Stop time of the interval in seconds</td>
   <tr><td>INT_4U</td><td>step</td><td>Step size in seconds of the frame files.</td>
   </table>
*/
