#ifndef DISKCACHE_DIR_HASH_HH
#define DISKCACHE_DIR_HASH_HH
   
// System Header Files   
#include <string>   
#include <vector>   
#include <list>   
#include <map>   
   
// General Header Files   
#include "general/types.hh"
#include "general/SharedPtr.hh"
   
// Local Header Files   
#include "rwsingleton.hh"
#include "dirinfo.hh"
   
namespace diskCache
{
   // Forward declarations
   class CacheFile;
   class LogMessage;
   
   //
   //: Directory hash (singleton) class.
   //
   // This class is designed to keep track of all identified
   // directories (and their recursive subdirectories) under
   // MOUNT_PT directory list (regardless of either they contain 
   // frame data or not).
   //
   class DirectoryHash
   {
   public:

      //: Declare singleton part of the class.
      DECLARE_READ_WRITE_SINGLETON( DirectoryHash );

   
   public: 
     typedef DirectoryInfo::resulting_dirs_container_type
     resulting_dirs_container_type;

     typedef DirectoryInfo::excluded_dirs_container_type
     excluded_dirs_container_type;
     typedef DirectoryInfo::dir_container_type	dir_container_type;

      //: Predicate used to check if directory should be excluded (actual 
      //: directory name matches any one from a specified list)
      class IsExcluded 
	: public std::unary_function< std::string, bool >
      {
      public:
         
         IsExcluded( const dir_container_type& dirs );
      
         bool operator()( const std::string& name ) const;
      
      
      private:
   
         //: A list of known directory names to be excluded
         const dir_container_type mDirNames; 
      };      
   
   
      //: Predicate used to check if directory has a specific
      //: parent directory.
      class HasParent
	: public std::unary_function< std::string, bool >
      {
      public:
         
         HasParent( const std::string& parent );
         virtual ~HasParent();
      
         virtual bool operator()( const std::string& dir ) const;

      
      protected:
   
         //: Full path for parent directory 
         std::string mParent;      

      
      private:
         //: Length of parent directory path
         size_t mParentSize;
      };
      
   
      static const INT_4U               SCANNED_DIRECTORIES_RESET;

      inline static INT_4U ScannedDirectories( )
      {
         MutexLock      lock( m_scanned_directories_baton );

         return m_scanned_directories;
      }

      inline static void ScannedDirectories( INT_4U Value )
      {
         MutexLock      lock( m_scanned_directories_baton );

         if ( Value == SCANNED_DIRECTORIES_RESET )
         {
            m_scanned_directories = 0;
         }
         else
         {
            m_scanned_directories += Value;
         }
      }

     const INT_4S
     GetMountPointMTime( const std::string& MountPoint,
			 const dir_container_type& ExcludeDirs );

     void AddDirectory( General::SharedPtr< DirectoryInfo > DirInfo );

     void RemoveDirectory( const std::string& DirectoryName );

      //: Mutators
      const std::list< std::string >
      scanDirectory( const std::string& dir,
		     const std::string& mount_pt,
		     const dir_container_type& exclude_dirs,
		     LogMessage& errors,
		     const INT_4S new_mtime = -1 );
   
      int
      removeDirectory( const std::string& dir,
		       resulting_dirs_container_type& RemovedDirectories,
		       LogMessage& errors,
		       const bool& unregister = false );
   
      void removeDirectories();
   
      //-----------------------------------------------------------------
      /// \brief Modify Hash entries based on modified exclude list
      /// 
      /// \param[in] AddedToList
      ///     A collection of directories that have been added to
      ///     the list of directories to be ignored
      /// \param[in] RemovedFromList
      ///     A collection of directories that have been remmoved from
      ///     the list of directories to be ignored
      /// 
      /// \return
      ///     A collection of directories that have been removed
      ///     from the hash since they were referenced in the
      ///     AddedToList parameter.
      //-----------------------------------------------------------------
      resulting_dirs_container_type
      ExcludeDirectories( const excluded_dirs_container_type& AddedToList,
			  const excluded_dirs_container_type& RemovedFromList,
			  LogMessage& errors );

      const resulting_dirs_container_type
      excludeDirectories( const excluded_dirs_container_type& dirs,
			  LogMessage& errors );
   
   
      //: Writer - not that neat of an interface for thread safety reason:
      //: cannot expose DirectoryInfo* pointer to avoid race 
      //: condition: removeDirectory() vs. CacheFile << DirectoryInfo*
      void writeDirectory( const std::string& dirname, CacheFile& stream );   

      ///---------------------------------------------------------
      /// \brief Write the directories to stream
      ///---------------------------------------------------------
      void WriteDirectories( std::list< std::string >& Dirs,
			     CacheFile& Stream );
   
      //: Reader
      void readDirectories( CacheFile& stream );

   
     ~DirectoryHash( );

   private:
      // Useful typedefs
      typedef General::SharedPtr< DirectoryInfo > dir_data_type;
      typedef std::map< std::string, dir_data_type > container_type;
      typedef container_type::value_type value_type;
      typedef container_type::mapped_type mapped_type;
      typedef container_type::iterator iterator;
      typedef container_type::const_iterator const_iterator;
         
   
      // Mutators
      std::pair< iterator, bool >
      add_directory( mapped_type Info );

      const resulting_dirs_container_type
      exclude_directory( const std::string& dir,
			 LogMessage& errors );
   
   
      //: Map to store directory information for current MOUNT_PT list and
      //: all available subdirectories (use map to keep all directories sorted
      //: in lexicographical order)
      //: Hash of directory name <---> DirectoryInfo 
      container_type mAllDirectories;
   
      static MutexLock::lock_type       m_scanned_directories_baton;
      static INT_4U                     m_scanned_directories;
      static MutexLock::lock_type       m_write_queue_baton;

   };
}

   
#endif   
