#ifndef FRAME_CPP_INTERFACE__FR_DETECTOR_H
#define FRAME_CPP_INTERFACE__FR_DETECTOR_H

#include "general/types.hh"

#include "framec/FrameC.h"
#include "framec/Stream.h"

#define DETECTOR_PREFIX_LEN 2

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  /** -----------------------------------------------------------
   * \brief latitude as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_8 fr_detector_latitude_t;
  /** -----------------------------------------------------------
   * \brief longitude as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_8 fr_detector_longitude_t;
  /** -----------------------------------------------------------
   * \brief elevation as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_4 fr_detector_elevation_t;
  /** -----------------------------------------------------------
   * \brief x axis azimuth as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_4 fr_detector_azimuth_x_t;
  /** -----------------------------------------------------------
   * \brief y axis azimuth as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_4 fr_detector_azimuth_y_t;
  /** -----------------------------------------------------------
   * \brief x axis altitude as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_4 fr_detector_altitude_x_t;
  /** -----------------------------------------------------------
   * \brief y axis altitude as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_4 fr_detector_altitude_y_t;
  /** -----------------------------------------------------------
   * \brief x axis midpoint as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_4 fr_detector_midpoint_x_t;
  /** -----------------------------------------------------------
   * \brief y axis midpoint as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef REAL_4 fr_detector_midpoint_y_t;
  /** -----------------------------------------------------------
   * \brief local seasonal time as used in the fr_detector structure.
   * ------------------------------------------------------------ */
  typedef INT_4S fr_detector_localtime_t;

  struct fr_detector {
    /** ---------------------------------------------------------
     * \brief Instrument name
     * --------------------------------------------------------- */
    char*			name;
    /** ---------------------------------------------------------
     * \brief Channel prefix for this detector.
     * --------------------------------------------------------- */
    char			prefix[ DETECTOR_PREFIX_LEN + 1 ];
    /** ---------------------------------------------------------
     * \brief Detector vertex latitude.
     * --------------------------------------------------------- */
    fr_detector_latitude_t	latitude;
    /** ---------------------------------------------------------
     * \brief Detector vertex longitude.
     * --------------------------------------------------------- */
    fr_detector_longitude_t	longitude;
    /** ---------------------------------------------------------
     * \brief Vertex elevation (meters)
     * --------------------------------------------------------- */
    fr_detector_elevation_t	elevation;
    /** ---------------------------------------------------------
     * \brief Orientation of X arm.
     * --------------------------------------------------------- */
    fr_detector_azimuth_x_t	azimuth_x;
    /** ---------------------------------------------------------
     * \brief Orientation of Y arm.
     * --------------------------------------------------------- */
    fr_detector_azimuth_y_t	azimuth_y;
    /** ---------------------------------------------------------
     * \brief Altitude angle of X arm.
     * --------------------------------------------------------- */
    fr_detector_altitude_x_t	altitude_x;
    /** ---------------------------------------------------------
     * \brief Altitude angle of X arm.
     * --------------------------------------------------------- */
    fr_detector_altitude_y_t	altitude_y;
    /** ---------------------------------------------------------
     * \brief Midpoint of the X cavity.
     * --------------------------------------------------------- */
    fr_detector_midpoint_x_t	midpoint_x;
    /** ---------------------------------------------------------
     * \brief Midpoint of the Y cavity.
     * --------------------------------------------------------- */
    fr_detector_midpoint_y_t	midpoint_y;
    /** ---------------------------------------------------------
     * \brief Local seasonal time.
     * --------------------------------------------------------- */
    fr_detector_localtime_t	localtime;

    struct fr_stat_data*	sdata;
    struct fr_detector*		next;
  };


  typedef struct fr_detector fr_detector_t;

  /** -----------------------------------------------------------
   * \brief Allocate a new Detector structure.
   *
   * \param[in] Error
   * If an error occurs, this will contain the details of the error.
   * For full details on error processing, please refer to FrameCError.
   *
   * \param[in] name
   * Instrument name as described in Appendix D of the frame spec.
   *
   * \param[in] prefix
   * Channel prefix for this detector as described in Appendix D.
   *
   * \param[in] latitude
   * Detector vertex latitude, geographical coordinates.
   *
   * \param[in] longitude
   * Detector vertex longitude, geographical coordinates.
   *
   * \param[in] elevation
   * Vertex elevation, meters, relative to WGS84 ellipsoid.
   *
   * \param[in] azimuth_x
   * Orientations of X arm, measured in radians East of North.
   *
   * \param[in] azimuth_y
   * Orientations of Y arm, measured in radians East of North.
   *
   * \param[in] altitude_x
   * Altitude (pitch) angle of X arm, measured in radians.
   *
   * \param[in] altitude_y
   * Altitude (pitch) angle of Y arm, measured in radians.
   *
   * \param[in] midpoint_x
   * Distance between the detector vertex and middle of the X cavity (meters).
   *
   * \param[in] midpoint_y
   * Distance between the detector vertex and middle of the Y cavity (meters).
   *
   * \param[in] localtime
   * Local seasonal time - UTC in seconds.
   *
   * \return
   * Upon successful completion, a newly allocated fr_detector_t
   * is returned.
   * It is the responsibility of the user to call FrDetectorFree
   * to recycle this resource.
   * ----------------------------------------------------------- */
  fr_detector_t* FrDetectorAlloc( FrameCError** Error,
				  const char* restrict name,
				  const char* restrict prefix,
				  fr_detector_latitude_t latitude,
				  fr_detector_longitude_t longitude,
				  fr_detector_elevation_t elevation,
				  fr_detector_azimuth_x_t azimuth_x,
				  fr_detector_azimuth_y_t azimuth_y,
				  fr_detector_altitude_x_t altitude_x,
				  fr_detector_altitude_y_t altitude_y,
				  fr_detector_midpoint_x_t midpoint_x,
				  fr_detector_midpoint_y_t midpoint_y,
				  fr_detector_localtime_t localtime);

  /** -----------------------------------------------------------
   * \brief Recyle the detector structure.
   *
   * \param[out] Error
   * If an error occurs, this will contain the details of the error.
   * For full details on error processing, please refer to FrameCError.
   *
   * \param[in] detector
   * The detector resouce to be recycled.
   * ----------------------------------------------------------- */
  void FrDetectorFree( FrameCError** Error,
		       fr_detector_t* detector );

  /** -----------------------------------------------------------
   * \brief Read a new Detector structure.
   *
   * \param[in] Error
   * If an error occurs, this will contain the details of the error.
   * For full details on error processing, please refer to FrameCError.
   *
   * \param[in] Stream
   * The stream from which to read the detector.
   *
   * \param[in] name
   * Instrument name.
   *
   * \return
   * Upon successful completion, a newly allocated fr_detector_t
   * is returned.
   * It is the responsibility of the user to call FrDetectorFree
   * to recycle this resource.
   * ----------------------------------------------------------- */
  fr_detector_t* FrDetectorRead( FrameCError** Error,
				 fr_file_t* restrict stream,
				 const char* restrict name );

  /** -----------------------------------------------------------
   * \brief Write a Detector structure to the stream.
   *
   * \param[in] Error
   * If an error occurs, this will contain the details of the error.
   * For full details on error processing, please refer to FrameCError.
   *
   * \param[in] Stream
   * The stream on which to write the detector.
   *
   * \return
   * Upon successful completion, a newly allocated fr_detector_t
   * is returned.
   * It is the responsibility of the user to call FrDetectorFree
   * to recycle this resource.
   * ----------------------------------------------------------- */
  int FrDetectorWrite( FrameCError** Error,
		       fr_file_t* restrict stream,
		       const fr_detector_t* restrict detector );

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */

#endif /* FRAME_CPP_INTERFACE__FR_DETECTOR_H */
