#ifndef FRAME_CPP_INTERFACE__FR_FRAME_H
#define FRAME_CPP_INTERFACE__FR_FRAME_H

#include "framecpp/config.h"

#include "general/types.hh"

#include "framec/FrameC.h"
#include "framec/Stream.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

  typedef INT_4S fr_frame_run_t;
  typedef INT_4U fr_frame_frame_t;
  typedef INT_4U fr_frame_dq_t;
  typedef gpstime_t fr_frame_t0_t;
  typedef INT_2U fr_frame_dt_t;

  typedef INT_2U fr_frame_offset_t;

  struct fr_frame {
    /**-----------------------------------------------------------------
     * \brief Name of project or other experiment description
     *
     * Examples of common values are, GEO, LIGO, VIRGO, and TAMA.
     *------------------------------------------------------------------*/
    char*		name;
    /**-----------------------------------------------------------------
     * \brief Run number
     *
     * This is the run number of the experiment and is monotonic.
     * Values less than zero are to be used when the data in the frame
     *   has been simultated.
     *------------------------------------------------------------------*/
    fr_frame_run_t	run;
    /**-----------------------------------------------------------------
     * \brief Frame number
     *
     * The frame number is monotonically increasing until end of run.
     * At the start of each run, the frame number is reset to zero.
     *------------------------------------------------------------------*/
    fr_frame_frame_t	frame;
    /**-----------------------------------------------------------------
     * \brief Data quality
     *
     * A logical word to denote top level quality of data.
     * Lowest order birs are reserved in pairs for the various
     * gravitational wave detectors.
     *------------------------------------------------------------------*/
    fr_frame_dq_t	dq;
    /**-----------------------------------------------------------------
     * \brief Frame start time.
     *------------------------------------------------------------------*/
    fr_frame_t0_t	t0;
    /**-----------------------------------------------------------------
     * \brief Frame length in seconds
     *------------------------------------------------------------------*/
    fr_frame_dt_t	dt;
  };

  typedef struct fr_frame fr_frame_t;

  /**-------------------------------------------------------------------
   * \brief Allocate a fr_frame structure.
   *
   * \note
   * This method should be used as there may be additional data allocated
   * to properly interface to the C++ layer.
   *
   * \param[out] Error
   *     If an error occurs, Error will be non-NULL and a detailed error
   *     message will be available.
   * \param[in] Name
   *     Name of project.
   * \param[in] Start
   *     GPS start time of frame.
   * \param[in] Dt
   *     Duration of frame.
   * \param[in] FrameNumber
   *     Frame number since start of run.
   *
   * \return
   *     Pointer to newly allocated fr_frame_t.
   *     If a failure happens, a NULL pointer will be returned.
   *--------------------------------------------------------------------*/
  fr_frame_t *FrFrameAlloc( FrameCError** Error,
			    const char *Name,
			    fr_frame_t0_t Start,
			    fr_frame_dt_t Dt,
			    fr_frame_frame_t FrameNumber );

  int FrFrameFree( FrameCError** Error,
		   fr_frame_t *frame );

  fr_frame_t *FrFrameRead( FrameCError** Error,
			   fr_file_t *restrict stream,
			   fr_frame_offset_t pos );

  int FrFrameWrite( FrameCError** Error,
		    fr_file_t *restrict stream,
		    fr_frame_t *restrict frame );

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */

#endif /* FRAME_CPP_INTERFACE__FR_FRAME_H */
