#ifndef FRAMECPP__COMMON__FrameStream_HH
#define FRAMECPP__COMMON__FrameStream_HH

#include "framecpp/config.h"

#include <list>
#include <memory>
#include <set>
#include <stdexcept>

#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/FrameSpec.hh"
#include "framecpp/Common/FrHeader.hh"
#include "framecpp/Common/FrStatData.hh"
#include "framecpp/Common/FrTOC.hh"

#include "framecpp/Common/IOStream.hh"

namespace FrameCPP
{
  namespace Common
  {
    class IStream;
    class OStream;
    class StreamRefInterface;
    class FrStatData;
    class FrDetector;
    class Verify;

    //===================================================================
    //===================================================================
    class FrameStream
      : virtual protected StreamBase
    {
    public:
      typedef FrameSpec::version_type version_type;

      version_type Version( ) const;

      const std::string& GetLastError( ) const;

      const MD5Sum& GetMD5Sum( ) const;

    protected:
      bool	m_closed_state;

      FrameStream( version_type Version, bool AllowZeroVersion );

      using StreamBase::frameSpecInfo;

      virtual void frameSpecInfo( const FrameSpec::Info* Spec );

      void setLastError( const std::string& Message );

      const StreamRefInterface& streamRef( ) const;

      void version_stream( version_type Version );

    protected:
      version_type memoryVersion( ) const;
      version_type streamVersion( ) const;

      virtual void reset( );

      std::auto_ptr< MD5SumFilter >	m_md5sumFilter;

    private:
      /// \brief Version number to use for frames in memory
      version_type			m_memory_version;

      /// \brief Version number to use for frames in stream
      version_type			m_stream_version;

      const StreamRefInterface*		m_stream_ref;

      std::string			m_last_error_message;
    };

    inline const std::string& FrameStream::
    GetLastError( ) const
    {
      return m_last_error_message;
    }

    inline const MD5Sum&  FrameStream::
    GetMD5Sum( ) const
    {
      MD5Sum* result( static_cast< MD5Sum* >( m_md5sumFilter.get( ) ) );
      if ( result == (MD5Sum*)NULL )
      {
	throw std::runtime_error( "No md5sum being calculated for frame" );
      }

      return *result;
    }

    inline void FrameStream::
    setLastError( const std::string& Message)
    {
      m_last_error_message = Message;
    }

    inline const StreamRefInterface& FrameStream::
    streamRef( ) const
    {
      if ( m_stream_ref )
      {
	return *m_stream_ref;
      }
      throw std::invalid_argument( "m_stream_ref is NULL" );
    }

    inline FrameStream::version_type FrameStream::
    Version( ) const
    {
      return m_memory_version;
    }

    inline FrameStream::version_type FrameStream::
    memoryVersion( ) const
    {
      return m_memory_version;
    }

    inline FrameStream::version_type FrameStream::
    streamVersion( ) const
    {
      return m_stream_version;
    }

    inline void FrameStream::
    version_stream( version_type Version )
    {
      m_stream_version = Version;
      if ( m_memory_version == 0 )
      {
	m_memory_version = m_stream_version;
      }
    }

    //===================================================================
    //===================================================================
    class IFrameStream
      : public FrameStream,
        protected IStream
    {
    public:
      typedef IStream::buffer_type	buffer_type;

      using IStream::GetCheckSumFile;
      using IStream::GetCheckSumFrame;
      using IStream::SetCheckSumFrame;
      using IStream::GetCheckSumObject;
      using IStream::SetCheckSumObject;

      IFrameStream( buffer_type* Stream, INT_2U Version = FRAME_SPEC_CURRENT );

      virtual void Cleanup( );

      const FrHeader& GetFrHeader( ) const;

      const FrTOC* GetTOC( ) const;

      INT_4U GetNumberOfFrames( ) const;

      FrameSpec::Info::verification_func_type VerificationFunc( ) const;

      FrameSpec::Object* Read( bool OwnObject = true );

      FrameSpec::Object* ReadNextFrameAsObject( );

      //-----------------------------------------------------------------
      /// \brief Read the detector via the table of contents
      ///
      /// \param[in] Name
      ///     The name of the detector to be read
      ///
      /// \return
      ///     Upon successfully reading the detector, a reference to
      ///     the detector is returned, otherwise an exception is
      ///     is thrown.
      ///
      /// \note
      ///     The caller is responsible to recycle the resources
      ///     incurred from this call.
      //-----------------------------------------------------------------
      std::auto_ptr< FrameSpec::Object > ReadDetector( const std::string& Name );

      std::auto_ptr< FrameSpec::Object > ReadFrame( bool Decompress = false );

      std::auto_ptr< FrameSpec::Object > ReadFrameN( INT_4U FrameIndex,
						     bool Decompress = false );

      std::auto_ptr< FrameSpec::Object > ReadFrStatData( INT_8U Position );

      INT_4U
      ReadFrStatData( const std::string& NamePattern,
		      const std::string& DetectorPattern,
		      const General::GPSTime& StartTime,
		      const General::GPSTime& EndTime,
		      const INT_4U Version,
		      FrStatData::Query& QueryResults );

      void SetAutoDeleteBuffer( const bool AutoDelete );

      void SetCheckSumFile( const CheckSum::kind_type Type );

      void SetMD5Sum( bool Value );

      void VerifyHeader( Verify& Verifier );

      using IStream::good;
      using IStream::eof;

    protected:
      //-----------------------------------------------------------------
      //
      //-----------------------------------------------------------------
      virtual std::auto_ptr< PTR_STRUCT_BASE > Next( FrameSpec::Object* Obj );

      //-----------------------------------------------------------------
      // Routines dealing with specific data types
      //-----------------------------------------------------------------
      std::auto_ptr< FrameSpec::Object >
      advanceToFrame( INT_4U Frame );

      FrameSpec::Object*
      readFrameHSubset( INT_4U Frame, INT_4U ElementMask );

      FrameSpec::Object*
      readFrAdcData( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrAdcData( INT_4U Frame, INT_4U Channel );

      FrameSpec::Object*
      readFrAdcStruct( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrAdcStruct( INT_4U Frame, INT_4U Channel );

      FrameSpec::Object*
      readFrEvent( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrEventStruct( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrProcData( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrProcData( INT_4U Frame, INT_4U Channel );

      FrameSpec::Object*
      readFrProcStruct( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrProcStruct( INT_4U Frame, INT_4U Channel );

      FrameSpec::Object*
      readFrRawData( INT_4U Frame );

      FrameSpec::Object*
      readFrSerData( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrSerStruct( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrSimData( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrSimStruct( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrSimEvent( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readFrSimEventStruct( INT_4U Frame, const std::string& Channel );

      FrameSpec::Object*
      readNextFrame( );

      std::auto_ptr< StreamRefInterface >
      readStreamRef( );

      void readTOC( );

      virtual void reset( );

      //-----------------------------------------------------------------
      // General reading routines
      //-----------------------------------------------------------------
      std::auto_ptr< FrameSpec::Object >
      readObject( const StreamRefInterface* StreamRef );

      void readRecursive( INT_4U StopOnClassId );

      bool removeRecursiveDependant( const StreamRefInterface* StreamRef );

      //-----------------------------------------------------------------
      // Other functions
      //-----------------------------------------------------------------
      virtual void addRecursiveDependant( const PTR_STRUCT_BASE* Ref );

      void addRecursiveDependant( const StreamRefInterface* StreamRef );

      virtual void pushResolver( resolver* Resolver );

    private:
      struct detector_info_type {
	INT_2U	m_class;
	INT_4U	m_instance;
      };

      typedef std::map< INT_8U, detector_info_type> detector_pos_type;

      std::auto_ptr< FrHeader >	m_file_header;
      std::auto_ptr< FrTOC >	m_toc;
      bool			m_toc_loaded;
      /// state variable for reading frame subset.
      bool			m_reading_frame_subset;
      //-----------------------------------------------------------------
      /// Set of references which need to resolved.
      /// Used by recursive reading function.
      //-----------------------------------------------------------------
      std::set< INT_8U >	m_to_be_resolved;
      bool			m_log_recursive_dependants;
      //-----------------------------------------------------------------
      /// \brief Current frame index
      //-----------------------------------------------------------------
      INT_4U			m_next_frame_index;

      detector_pos_type		m_detector_pos;

      //-----------------------------------------------------------------
      /// \brief Load table of contents from stream
      //-----------------------------------------------------------------
      void load_toc( );

      //-----------------------------------------------------------------
      /// \brief Modify an object so it is appropriate for the stream.
      ///
      /// \param[in] Obj
      ///     The object that should be transformed.
      ///
      /// \return
      ///     A pointer to the object appropriately converted.
      ///     If no conversion was necessary, the same pointer is
      ///     returned.
      //-----------------------------------------------------------------
      FrameSpec::Object* morph( FrameSpec::Object* Obj );

      template< INT_4U ClassType, typename ChannelType >
      FrameSpec::Object*
      read_fr_struct( INT_4U Frame, ChannelType Channel,
		      bool Recursively );

      template< INT_4U ClassType, typename ChannelType >
      std::auto_ptr< FrameSpec::Object >
      advance_to_fr_struct( INT_4U FrameIndex, ChannelType Channel );

      template< INT_4U ClassType, typename ChannelType >
      INT_8U position( const FrTOC* TOC, INT_4U Frame, const ChannelType& Channel ) const;
    };

    inline FrameSpec::Info::verification_func_type IFrameStream::
    VerificationFunc( ) const
    {
      return frameSpecInfo( ).VerificationFunc( );
    }
    
    inline const FrHeader& IFrameStream::
    GetFrHeader( ) const
    {
      if ( m_file_header.get( ) )
      {
	return *(m_file_header.get( ) );
      }
      std::ostringstream	msg;
      
      msg << "FrHeader has not been initialized yet in this stream"
	;
      throw std::runtime_error( msg.str( ) );
    }

    inline INT_4U IFrameStream::
    GetNumberOfFrames( ) const
    {
      return GetTOC( )->nFrame( );
    }

    inline const FrTOC* IFrameStream::
    GetTOC( ) const
    {
      if ( m_toc_loaded == false )
      {
	const_cast< IFrameStream* >( this )->load_toc( );
      }
      if ( m_toc.get( ) )
      {
	return m_toc.get( );
      }
      throw
	std::runtime_error( "No Table of Contents object for Stream" );
    }


    inline void IFrameStream::
    SetAutoDeleteBuffer( const bool AutoDelete )
    {
      setAutoDeleteBuffer( AutoDelete );
    }

    inline void IFrameStream::
    SetCheckSumFile( const CheckSum::kind_type Type )
    {
      CheckSumFilter*	fcs = GetCheckSumFile( );
      if ( fcs
	   && ( fcs->Type( )  == Type ) )
      {
	return;
      }
      IStream::SetCheckSumFile( Type );
      fcs = GetCheckSumFile( );
      if ( fcs )
      {
	if ( cmn_streamsize_type( tellg( ) ) != m_file_header->Bytes( ) )
	{
	  throw
	    std::logic_error( "Unable to set file checksum: position error" );
	}
	m_file_header->Filter( *fcs );
      }
    }

    inline FrameSpec::Object* IFrameStream::
    ReadNextFrameAsObject( )
    {
      return readNextFrame( );
    }

    inline void IFrameStream::
    SetMD5Sum( bool Value )
    {
      if ( m_md5sumFilter.get( ) )
      {
	rdbuf( )->FilterRemove( m_md5sumFilter.get( ) );
      }
      if ( Value )
      {
	if ( m_md5sumFilter.get( ) )
	{
	  m_md5sumFilter->Reset( );
	}
	else
	{
	  m_md5sumFilter.reset( new MD5SumFilter( ) );
	}
	rdbuf( )->FilterAdd( m_md5sumFilter.get( ) );
      }
      else
      {
	m_md5sumFilter.reset( (MD5SumFilter*)NULL );
      }
    }

    //===================================================================
    //===================================================================
    class OFrameStream
      : public FrameStream,
	protected OStream
    {
    public:
      typedef OStream::buffer_type	buffer_type;
      typedef INT_4U			chkSum_type;

      using OStream::SetCheckSumFile;
      using OStream::good;

      OFrameStream( buffer_type* Buffer,
		    INT_2U Version = FRAME_SPEC_CURRENT );

      ~OFrameStream( );

      void Close( );

      virtual INT_4U GetNumberOfFrames( ) const;

      virtual INT_8U GetTOCOffset( ) const;

      void SetMD5Sum( bool Value );

      void Write( const FrameSpec::Object* Obj );

      /// \brief Write Frame data to the stream
      void WriteFrame( const FrameSpec::Object& FrameObject,
		       chkSum_type FrameChecksumType
		       = Common::CheckSum::NONE );

      /// \brief Write Frame data to the stream
      void WriteFrame( const FrameSpec::Object& FrameObject,
		       INT_2U CompressionScheme,
		       INT_2U CompressionLevel,
		       chkSum_type FrameChecksumType = CheckSum::NONE );

      void WriteFrStatData( const FrameSpec::Object& FrameObject );

    private:
      typedef std::set< INT_2U > 		defined_container_type;
      typedef std::list< const FrStatData* >	fr_stat_list_type;
      typedef std::map< const FrDetector*, fr_stat_list_type >
      fr_stat_container_type;

      std::auto_ptr< FrTOC >	m_toc;
      INT_8U			m_toc_location;
      INT_4U			m_frame_count;
      defined_container_type	m_defined;
      fr_stat_container_type	m_fr_stats;
      bool			m_has_header;

      void write_stream_objects( );
    };

    inline INT_4U OFrameStream::
    GetNumberOfFrames( ) const
    {
      return m_frame_count;
    }

    inline INT_8U OFrameStream::
    GetTOCOffset( ) const
    {
      return m_toc_location;
    }

    inline void OFrameStream::
    SetMD5Sum( bool Value )
    {
      if ( m_md5sumFilter.get( ) )
      {
	rdbuf( )->FilterRemove( m_md5sumFilter.get( ) );
      }
      if ( Value )
      {
	if ( m_md5sumFilter.get( ) )
	{
	  m_md5sumFilter->Reset( );
	}
	else
	{
	  m_md5sumFilter.reset( new MD5SumFilter( ) );
	}
	rdbuf( )->FilterAdd( m_md5sumFilter.get( ) );
      }
      else
      {
	m_md5sumFilter.reset( (MD5SumFilter*)NULL );
      }
    }

  }
}

#endif /* FRAMECPP__COMMON__FrameStream_HH */
